use crate::name_analysis::symbol::Symbol;
use std::collections::HashMap;
use std::fmt::Display;
use crate::name_analysis::symbol_table::SymbolInsertError::SymbolAlreadyDefined;
use crate::name_analysis::symbol_table::SymbolLookupError::NoDefinition;

pub struct Scope {
    parent: Option<usize>,
    function_and_variable_symbols: HashMap<String, Symbol>,
    type_and_module_symbols: HashMap<String, Symbol>,
    debug_name: String,
}

impl Scope {
    pub fn new(parent: Option<usize>, debug_name: String) -> Scope {
        Scope {
            parent,
            function_and_variable_symbols: HashMap::new(),
            type_and_module_symbols: HashMap::new(),
            debug_name,
        }
    }
}

pub enum SymbolInsertError {
    SymbolAlreadyDefined(Symbol),
}

pub enum SymbolLookupError {
    NoDefinition
}

pub struct SymbolTable {
    scopes: Vec<Scope>,
    current_scope_id: usize,
}

/// Contains a vec of scopes, like a flattened tree
impl SymbolTable {
    pub fn new() -> Self {
        let mut t = SymbolTable {
            scopes: vec![Scope::new(None, String::from("GlobalScope"))],
            current_scope_id: 0,
        };
        t
    }

    pub fn current_scope_id(&self) -> usize {
        self.current_scope_id
    }

    pub fn push_scope(&mut self, debug_name: &str) {
        let id = self.scopes.len();
        self.scopes
            .push(Scope::new(Some(self.current_scope_id), debug_name.to_string()));
        self.current_scope_id = id;
    }

    pub fn pop_scope(&mut self) {
        if let Some(parent_id) = self.scopes[self.current_scope_id].parent {
            self.current_scope_id = parent_id;
        }
    }

    pub fn insert(&mut self, name: &str, symbol: Symbol) -> Result<(), SymbolInsertError> {
        match symbol {
            Symbol::Function(_) | Symbol::Variable(_) => {
                self.insert_function_or_variable(name, symbol)
            }
            Symbol::Module(_) => self.insert_module_or_type(name, symbol),
        }
    }

    fn insert_function_or_variable(&mut self, name: &str, symbol: Symbol) -> Result<(), SymbolInsertError> {
        if let Some(defined_symbol) = self.scopes[self.current_scope_id]
            .function_and_variable_symbols
            .get(name)
        {
            Err(SymbolAlreadyDefined(defined_symbol.clone()))
        } else {
            self.scopes[self.current_scope_id]
                .function_and_variable_symbols
                .insert(name.to_string(), symbol);
            Ok(())
        }
    }

    fn insert_module_or_type(&mut self, name: &str, symbol: Symbol) -> Result<(), SymbolInsertError> {
        if let Some(defined_symbol) = self.scopes[self.current_scope_id]
            .type_and_module_symbols
            .get(name)
        {
            Err(SymbolAlreadyDefined(defined_symbol.clone()))
        } else {
            self.scopes[self.current_scope_id]
                .type_and_module_symbols
                .insert(name.to_string(), symbol);
            Ok(())
        }
    }

    pub fn lookup(&self, name: &str, scope_id: usize) -> Result<&Symbol, SymbolLookupError> {
        let mut scope_opt = Some(&self.scopes[scope_id]);
        while let Some(scope) = scope_opt {
            if let Some(symbol) = scope.function_and_variable_symbols.get(name) {
                return Ok(symbol);
            }
            scope_opt = if let Some(parent_id) = scope.parent {
                Some(&self.scopes[parent_id])
            } else {
                None
            };
        }
        Err(NoDefinition)
    }
}

impl Display for SymbolTable {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        writeln!(f, "SymbolTable(current_scope = {})", self.current_scope_id)?;
        for (i, scope) in self.scopes.iter().enumerate() {
            writeln!(f, "Scope {} {}", i, scope.debug_name)?;
            for (name, symbol) in &scope.function_and_variable_symbols {
                writeln!(f, "  {}({})", name, symbol)?;
            }
        }
        Ok(())
    }
}
